﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.IO;
using System.Text.RegularExpressions;
using System.Globalization;
using NiVE2.Plugin.Interface;
using NiVE2.Drawing;
using NiVE2.Utils;
using NiVE2.Utils.Cache;

namespace ruche.nive2.io
{
    /// <summary>
    /// フレーム番号指定の画像ファイル群を動画として読み込むNiVE2入力プラグインクラス。
    /// </summary>
    [Serializable]
    public class FrameImageInput : InputChild
    {
        /// <summary>
        /// 既定のフレームレート。
        /// </summary>
        private const double DefaultFrameRate = 30;

        /// <summary>
        /// フレーム数字スタイル。
        /// </summary>
        private const NumberStyles FrameNumberStyles =
            NumberStyles.AllowLeadingWhite | NumberStyles.AllowTrailingWhite;

        /// <summary>
        /// ファイル名正規表現。
        /// </summary>
        private static readonly Regex FileNameRegex = new Regex(
            @"^(.*?)(\d+)\s*(?:\[\s*(\d+)\s*\]\s*)?$",
            RegexOptions.Compiled);

        /// <summary>
        /// 最初に読み込んだファイルのパス。
        /// </summary>
        private string _filePath = null;

        /// <summary>
        /// ベースディレクトリパス。
        /// </summary>
        private string _dirPath = ".";

        /// <summary>
        /// メディア名。
        /// </summary>
        private string _mediaName = null;

        /// <summary>
        /// メディアの長さ。
        /// </summary>
        private double _length = 0;

        /// <summary>
        /// メディアのフレームレート。
        /// </summary>
        private double _frameRate = DefaultFrameRate;

        /// <summary>
        /// メディアのサイズ。
        /// </summary>
        private Size _imageSize = new Size();

        /// <summary>
        /// フレーム位置と対応するファイル名のハッシュ配列。
        /// </summary>
        private Dictionary<int, string> _frameFileNames = null;

        /// <summary>
        /// コンストラクタ。
        /// </summary>
        public FrameImageInput()
        {
            // 何もしない
        }

        /// <summary>
        /// ファイル名を接頭辞とフレーム位置、およびある場合はフレームレートに
        /// 分割する。
        /// </summary>
        /// <param name="name">拡張子なしのファイル名。</param>
        /// <param name="nameHead">接頭辞の設定先。</param>
        /// <param name="frame">フレーム位置の設定先。</param>
        /// <param name="frameRate">フレームレートの設定先。</param>
        /// <returns>分割できたならば true 、そうでなければ false 。</returns>
        private bool SplitFileName(
            string name,
            out string nameHead,
            out int frame,
            out double frameRate)
        {
            // フォーマット1: nameHead + frame
            // フォーマット2: nameHead + frame + '[' + frameRate + ']'
            bool result = false;

            Match match = FileNameRegex.Match(name);
            if (match.Success)
            {
                result = true;

                nameHead = match.Groups[1].Value;
                frame = int.Parse(match.Groups[2].Value);
                string frVal = match.Groups[3].Value;
                if (string.IsNullOrEmpty(frVal))
                {
                    frameRate = DefaultFrameRate;
                }
                else
                {
                    frameRate = double.Parse(frVal);
                    if (frameRate <= 0)
                    {
                        result = false;
                    }
                }
            }
            else
            {
                nameHead = null;
                frame = 0;
                frameRate = 0.0;
            }

            return result;
        }

        #region InputChild メンバ

        public override int ChildCode
        {
            get { return 0; }
        }

        public override string MediaName
        {
            get { return _mediaName; }
        }

        public override bool HaveVideo
        {
            get { return true; }
        }

        public override bool HaveAudio
        {
            get { return false; }
        }

        public override bool IsImage
        {
            get { return false; }
        }

        public override double Length
        {
            get { return _length; }
        }

        public override InputBase Parent
        {
            get { return this; }
        }

        public override MediaInfo GetMedaiInfo()
        {
            return new MediaInfo(
                Length, _frameRate, true, false, false, _imageSize);
        }

        public override PropertyBase[] GetDefaultProperty()
        {
            return new PropertyBase[] { };
        }

        public override PropertyEditControlBase[] GetControl()
        {
            return new PropertyEditControlBase[] { };
        }

        public override RectangleF GetBounds(
            double time,
            IComposition composition,
            ReadOnlyDictionary<string, PropertyBase> property)
        {
            return new RectangleF(
                0,
                0,
                (float)(_imageSize.Width * composition.ResolutionRate),
                (float)(_imageSize.Height * composition.ResolutionRate));
        }

        public override NBitmap GetImage(
            double time,
            IComposition composition,
            ReadOnlyDictionary<string, PropertyBase> property)
        {
            NBitmap result = null;

            double resoRate = composition.ResolutionRate;
            Size resoSize = new Size(
                (int)(_imageSize.Width * resoRate),
                (int)(_imageSize.Height * resoRate));

            long orgCacheSize = _imageSize.Width * _imageSize.Height * 4;
            long resoCacheSize = resoSize.Width * resoSize.Height * 4;
            long tempCacheSize = orgCacheSize + resoCacheSize;

            try
            {
                // 内部のFPSでフレーム位置計算
                for (int frame = (int)(time * _frameRate); frame >= 0; --frame)
                {
                    // 取得ファイル決定
                    string fileName = null;
                    if (_frameFileNames.TryGetValue(frame, out fileName))
                    {
                        // 画像ファイル読み込み＆出力サイズ画像作成
                        using (CacheLocker cacheLock = new CacheLocker(tempCacheSize))
                        using (Bitmap temp = new Bitmap(_dirPath + @"\" + fileName))
                        using (
                            Bitmap canvas = new Bitmap(
                                resoSize.Width,
                                resoSize.Height,
                                PixelFormat.Format32bppArgb))
                        {
                            // 出力サイズ画像へ描画
                            using (Graphics g = Graphics.FromImage(canvas))
                            {
                                g.CompositingMode = CompositingMode.SourceCopy;
                                g.CompositingQuality =
                                    CompositingQuality.HighQuality;
                                g.InterpolationMode =
                                    InterpolationMode.NearestNeighbor;
                                g.ScaleTransform(
                                    (float)resoRate,
                                    (float)resoRate,
                                    MatrixOrder.Append);
                                g.DrawImage(temp, 0, 0, temp.Width, temp.Height);
                            }

                            // NBitmap 作成
                            CacheManager.Manager.DemandTemporalMemory(resoCacheSize);
                            result = new NBitmap(canvas);
                        }
                        break;
                    }
                }
            }
            catch { }
            finally
            {
                // 読み込めなかったならば空の画像を作成
                if (result == null)
                {
                    CacheManager.Manager.DemandTemporalMemory(resoCacheSize);
                    result = new NBitmap(resoSize.Width, resoSize.Height);
                }

                // 出力データを弱参照キャッシュに追加
                CacheManager.Manager.AddWeakCache(result, result.DataSize);
            }

            return result;
        }

        public override byte[] GetAudio(double time, int length, IComposition composition)
        {
            // 呼ばれないはず
            throw new NotImplementedException();
        }

        #endregion

        #region InputBase メンバ

        public override string CorrespondingType
        {
            get
            {
                return (
                    "すべての形式(*.bmp;*.jpg;*.gif;*.png;*.tiff)|" +
                    "*.bmp;*.jpg;*.gif;*.png;*.tiff|" +
                    "ビットマップ(*.bmp)|*.bmp|" +
                    "JPEG(*.jpg)|*.jpg|" +
                    "GIF(*.gif)|*.gif|" +
                    "PNG(*.png)|*.png|" +
                    "TIFF(*.tiff)|*.tiff");
            }
        }

        public override string FileName
        {
            get { return _filePath; }
        }

        public override void Initialize()
        {
            // 何もしない
        }

        public override bool Load(string file)
        {
            _filePath = file;

            try
            {
                // ディレクトリ、名前、拡張子に分ける
                _dirPath = Path.GetDirectoryName(file);
                string name = Path.GetFileNameWithoutExtension(file);
                string ext = Path.GetExtension(file);

                // 名前を接頭辞、開始フレーム、フレームレートに分ける
                string nameHead = null;
                int startFrame = 0;
                if (!SplitFileName(name, out nameHead, out startFrame, out _frameRate))
                {
                    return false;
                }

                // メディア名決定
                if (nameHead.Length > 0)
                {
                    _mediaName = nameHead;
                }
                else
                {
                    _mediaName = Path.GetFileName(_dirPath);
                    if (string.IsNullOrEmpty(_mediaName))
                    {
                        _mediaName = "FrameImages";
                    }
                }
                _mediaName += "[" + _frameRate + "fps]";

                // 画像サイズ決定
                using (CacheLocker cacheLock = new CacheLocker(1280 * 1024 * 4))
                using (Bitmap temp = new Bitmap(_filePath))
                {
                    _imageSize = temp.Size;
                }

                // ディレクトリ内ファイル取得
                string[] files =
                    Directory.GetFiles(_dirPath, nameHead + "*" + ext);

                // ハッシュ配列初期化
                _frameFileNames = new Dictionary<int, string>(files.Length);
                _frameFileNames.Add(startFrame, Path.GetFileName(file));

                // ファイル追加
                int endFrame = 0;
                foreach (string fp in files)
                {
                    if (string.Compare(Path.GetExtension(fp), ext, true) == 0)
                    {
                        string n = Path.GetFileNameWithoutExtension(fp);
                        string head = null;
                        int frame = 0;
                        double dummy = 0;
                        if (
                            SplitFileName(n, out head, out frame, out dummy) &&
                            head == nameHead &&
                            frame > startFrame &&
                            !_frameFileNames.ContainsKey(frame))
                        {
                            _frameFileNames.Add(frame, Path.GetFileName(fp));
                            endFrame = Math.Max(endFrame, frame);
                        }
                    }
                }

                if (_frameFileNames.Count <= 0)
                {
                    return false;
                }

                // 長さ決定
                _length = (endFrame - startFrame + 1) / _frameRate;
            }
            catch
            {
                return false;
            }

            return true;
        }

        public override InputChild[] GetChildInput()
        {
            return new InputChild[] { this };
        }

        public override object SaveInnerData()
        {
            return null;
        }

        public override void LoadInnerData(object data)
        {
            // 何もしない
        }

        #endregion

        #region PluginBase メンバ

        public override string PluginName
        {
            get { return "フレーム番号イメージ読み込み"; }
        }

        public override string Author
        {
            get { return "ルーチェ"; }
        }

        public override string InfoLink
        {
            get { return "http://www.ruche-home.net/"; }
        }

        public override string Description
        {
            get
            {
                return "フレーム番号指定の画像ファイル群を動画として読み込みます。";
            }
        }

        #endregion

        #region IDisposable メンバ

        public override void Dispose()
        {
            // 何もしない
        }

        #endregion
    }
}
